@@ -1,5 +1,3 @@
-from typing import List
-
 from nerblackbox.modules.ner_training.data_preprocessing.tools.bert_dataset import (
     BertDataset,
 )
@@ -9,15 +7,27 @@ from nerblackbox.modules.ner_training.data_preprocessing.tools.csv_reader import
 from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example import (
     InputExample,
 )
-from nerblackbox.modules.ner_training.data_preprocessing.tools.input_example_to_tensors import (
-    InputExampleToTensors,
+from nerblackbox.modules.ner_training.data_preprocessing.tools.input_examples_to_tensors import (
+    InputExamplesToTensors,
 )
+from nerblackbox.tests.utils import PseudoDefaultLogger
 from nerblackbox.modules.utils.util_functions import get_dataset_path
-from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
+from torch.utils.data import DataLoader, Sampler, RandomSampler, SequentialSampler
+from pkg_resources import resource_filename
+
+from typing import List, Dict, Tuple, Optional, Any
+
+InputExamples = List[InputExample]
 
 
 class DataPreprocessor:
-    def __init__(self, tokenizer, do_lower_case, default_logger, max_seq_length=64):
+    def __init__(
+        self,
+        tokenizer,
+        do_lower_case: bool,
+        default_logger: PseudoDefaultLogger,
+        max_seq_length: int = 64,
+    ):
         """
         :param tokenizer:      [transformers Tokenizer]
         :param do_lower_case:  [bool] if True, make text data lowercase
@@ -29,19 +39,24 @@ class DataPreprocessor:
         self.default_logger = default_logger
         self.max_seq_length = max_seq_length
 
-    def get_input_examples_train(self, dataset_name, prune_ratio):
+    def get_input_examples_train(
+        self, prune_ratio: Dict[str, float], dataset_name: Optional[str] = None
+    ) -> Tuple[Dict[str, InputExamples], List[str]]:
         """
-        get input examples for TRAIN from csv files
-        -------------------------------------------
-        :param dataset_name:     [str], e.g. 'suc'
-        :param prune_ratio:      [dict], e.g. {'train': 1.0, 'val': 1.0, 'test': 1.0} -- pruning ratio for data
-        :return: input_examples: [dict] w/ keys = 'train', 'val', 'test' & values = [list] of [InputExample]
-        :return: tag_list:       [list] of tags present in the dataset, e.g. ['O', 'PER', ..]
-        """
-        if prune_ratio is None:
-            prune_ratio = {"train": 1.0, "val": 1.0, "test": 1.0}
+        - get input examples for TRAIN from csv files
 
-        dataset_path = get_dataset_path(dataset_name)
+        Args:
+            dataset_name:     [str], e.g. 'suc'
+            prune_ratio:      [dict], e.g. {'train': 1.0, 'val': 1.0, 'test': 1.0} -- pruning ratio for data
+
+        Returns:
+            input_examples: [dict] w/ keys = 'train', 'val', 'test' & values = [list] of [InputExample]
+            tag_list:       [list] of tags present in the dataset, e.g. ['O', 'PER', ..]
+        """
+        if dataset_name is None:
+            dataset_path = resource_filename("nerblackbox", "tests/test_data")
+        else:
+            dataset_path = get_dataset_path(dataset_name)
 
         # csv_reader
         csv_reader = CsvReader(
@@ -63,12 +78,17 @@ class DataPreprocessor:
             csv_reader.tag_list
         )
 
-    def get_input_examples_predict(self, examples):
+    def get_input_examples_predict(
+        self, examples: List[str]
+    ) -> Dict[str, InputExamples]:
         """
-        get input examples for PREDICT from input argument examples
-        -----------------------------------------------------------
-        :param examples:         [list] of [str]
-        :return: input_examples: [dict] w/ key = 'predict' & value = [list] of [InputExample]
+        - get input examples for PREDICT from input argument examples
+
+        Args:
+            examples:       [list] of [str]
+
+        Returns:
+            input_examples: [dict] w/ key = 'predict' & value = [list] of [InputExample]
         """
         # create input_examples
         if self.do_lower_case:
@@ -78,8 +98,8 @@ class DataPreprocessor:
             "predict": [
                 InputExample(
                     guid="",
-                    text_a=example,
-                    tags_a=" ".join(
+                    text=example,
+                    tags=" ".join(
                         ["O" for _ in range(len(example.split()))]
                     ),  # pseudo tags
                 )
@@ -90,20 +110,27 @@ class DataPreprocessor:
 
         return input_examples
 
-    def to_dataloader(self, input_examples, tag_list, batch_size):
+    def to_dataloader(
+        self,
+        input_examples: Dict[str, InputExamples],
+        tag_list: List[str],
+        batch_size: int,
+    ) -> Dict[str, DataLoader]:
         """
-        turn input_examples into dataloader
-        -----------------------------------
-        :param input_examples:
-        :param input_examples: [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
-                                         values = [list] of [InputExample]
-        :param tag_list:         [list] of tags present in the dataset, e.g. ['O', 'PER', ..]
-        :param batch_size:       [int]
-        :return: _dataloader:    [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
-                                           values = [torch Dataloader]
+        - turn input_examples into dataloader
+
+        Args:
+            input_examples: [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
+                                      values = [list] of [InputExample]
+            tag_list:       [list] of tags present in the dataset, e.g. ['O', 'PER', ..]
+            batch_size:     [int]
+
+        Returns:
+            _dataloader:    [dict] w/ keys = ['train', 'val', 'test'] or ['predict'] &
+                                      values = [torch Dataloader]
         """
         # input_example_to_tensors
-        input_example_to_tensors = InputExampleToTensors(
+        input_examples_to_tensors = InputExamplesToTensors(
             self.tokenizer,
             max_seq_length=self.max_seq_length,
             tag_tuple=tuple(tag_list),
@@ -112,16 +139,31 @@ class DataPreprocessor:
 
         _dataloader = dict()
         for phase in input_examples.keys():
-            # dataloader
-            data = BertDataset(
-                input_examples[phase], transform=input_example_to_tensors
+            encodings = input_examples_to_tensors(
+                input_examples[phase], predict=phase == "predict"
             )
 
+            # dataloader
+            data = BertDataset(
+                encodings=encodings
+            )  # data[j] = 4 torch tensors corresponding to EncodingKeys
+            if self.default_logger:
+                self.default_logger.log_info(
+                    f"[after pre-preprocessing] {phase.ljust(5)} data: {len(data)} examples"
+                )
+
+            assert phase in [
+                "train",
+                "val",
+                "test",
+                "predict",
+            ], f"ERROR! phase = {phase} unknown."
+            sampler: Optional[Sampler]
             if phase == "train":
                 sampler = RandomSampler(data)
             elif phase in ["val", "test"]:
                 sampler = SequentialSampler(data)
-            else:
+            else:  # if phase == "predict":
                 sampler = None
 
             _dataloader[phase] = DataLoader(
@@ -133,7 +175,7 @@ class DataPreprocessor:
     ####################################################################################################################
     # HELPER
     ####################################################################################################################
-    def _prune_examples(self, list_of_examples, phase, ratio=None):
+    def _prune_examples(self, list_of_examples: List[Any], phase: str, ratio: float):
         """
         prunes list_of_examples by taking only the first examples
         ---------------------------------------------------------
@@ -142,13 +184,14 @@ class DataPreprocessor:
         :param (Optional) ratio: [float], e.g. 0.5
         :return: [list], e.g. of [InputExample]
         """
-        if ratio is None:
+        if ratio == 1.0:
             return list_of_examples
         else:
             num_examples_old = len(list_of_examples)
             num_examples_new = int(ratio * float(num_examples_old))
             info = f"> {phase.ljust(5)} data: use {num_examples_new} of {num_examples_old} examples"
-            self.default_logger.log_info(info)
+            if self.default_logger:
+                self.default_logger.log_info(info)
             return list_of_examples[:num_examples_new]
 
     @staticmethod
