@@ -1,9 +1,10 @@
-import warnings
+import os
+from os.path import join
 import numpy as np
-from seqeval.metrics import classification_report as classification_report_seqeval
-from sklearn.metrics import classification_report as classification_report_sklearn
 import pytorch_lightning as pl
 from abc import ABC, abstractmethod
+import torch
+from typing import List, Dict
 
 from transformers import AdamW
 from transformers import get_linear_schedule_with_warmup
@@ -15,10 +16,10 @@ from transformers import AutoTokenizer
 from nerblackbox.modules.ner_training.data_preprocessing.data_preprocessor import (
     DataPreprocessor,
 )
-from nerblackbox.modules.ner_training.metrics.ner_metrics import NerMetrics
-from nerblackbox.modules.ner_training.metrics.ner_metrics import convert_to_chunk
 from nerblackbox.modules.utils.util_functions import split_parameters
-from nerblackbox.modules.utils.env_variable import env_variable
+from nerblackbox.modules.ner_training.ner_model_evaluation import NerModelEvaluation
+
+NEWLINE_TOKENS = ["[newline]", "[NEWLINE]"]
 
 
 class NerModel(pl.LightningModule, ABC):
@@ -64,15 +65,35 @@ class NerModel(pl.LightningModule, ABC):
 
     def _preparations_data_general(self):
         """
-        :created attr: tokenizer         [transformers AutoTokenizer]
-        :created attr: data_preprocessor [DataPreprocessor]
+        :created attr: pretrained_model_name  [str]
+        :created attr: tokenizer              [transformers AutoTokenizer]
+        :created attr: data_preprocessor      [DataPreprocessor]
         :return: -
         """
-        # tokenizer
-        self.tokenizer = AutoTokenizer.from_pretrained(
-            self.pretrained_model_name, do_lower_case=False
-        )  # needs to be False !!
+        # 1. pretrained model name
+        try:
+            # use transformers model
+            AutoTokenizer.from_pretrained(
+                self.params.pretrained_model_name, use_fast=False
+            )
+            self.pretrained_model_name = self.params.pretrained_model_name
+        except ValueError:
+            # use local model
+            self.pretrained_model_name = join(
+                os.environ.get("DATA_DIR"),
+                "pretrained_models",
+                self.params.pretrained_model_name,
+            )
 
+        # 2. tokenizer
+        self.tokenizer = AutoTokenizer.from_pretrained(
+            self.pretrained_model_name,
+            do_lower_case=False,
+            additional_special_tokens=NEWLINE_TOKENS,
+            use_fast=True,
+        )  # do_lower_case needs to be False !!
+
+        # 3. data_preprocessor
         self.data_preprocessor = DataPreprocessor(
             tokenizer=self.tokenizer,
             do_lower_case=self.params.uncased,  # can be True !!
@@ -102,7 +123,7 @@ class NerModel(pl.LightningModule, ABC):
                     ii) _tag_ids_prediction_logits: [torch tensor] of shape [batch_size, seq_length, vocabulary_size]
         """
         return self.model(
-            _input_ids,
+            input_ids=_input_ids,
             attention_mask=_attention_mask,
             token_type_ids=_segment_ids,
             labels=_tag_ids,
@@ -118,7 +139,7 @@ class NerModel(pl.LightningModule, ABC):
         batch_train_loss = outputs[0]
 
         # logging
-        self.write_metrics_for_tensorboard("train", {"all+_loss": batch_train_loss})
+        self._write_metrics_for_tensorboard("train", {"all_loss": batch_train_loss})
 
         # debug
         if batch_idx == 0:
@@ -336,251 +357,64 @@ class NerModel(pl.LightningModule, ABC):
         """
         return _num_epochs * len(self.dataloader["train"])
 
-    ####################################################################################################################
-    # 2. VALIDATE / COMPUTE METRICS
-    ####################################################################################################################
-    def _validate_on_epoch(self, phase, outputs):
+    def _validate_on_epoch(
+        self, phase: str, outputs: List[List[torch.Tensor]]
+    ) -> Dict[str, float]:
         """
-        validate on all batches of one epoch, i.e. whole val or test dataset
-        --------------------------------------------------------------------
-        :param phase:   [str], 'val', 'test'
-        :param outputs: [list] of [list] w/ 3 elements [batch_loss, batch_tag_ids, batch_logits] for each batch
-        :return: [dict] w/ key '<phase>_loss' & value = mean batch loss [float]
+        Args:
+            phase:   [str], 'val', 'test'
+            outputs: [list] of [lists] = [batch_loss, batch_tag_ids, batch_logits] with 3 torch tensors for each batch
+
+        Returns:
+            [dict] w/ key '<phase>_loss' & value = mean batch loss [float]
         """
-        # to cpu/numpy
-        np_batch = {
-            "loss": [output[0].detach().cpu().numpy() for output in outputs],
-            "tag_ids": [
-                output[1].detach().cpu().numpy() for output in outputs
-            ],  # [batch_size, seq_length]
-            "logits": [
-                output[2].detach().cpu().numpy() for output in outputs
-            ],  # [batch_size, seq_length, num_tags]
-        }
-
-        # combine np_batch metrics to np_epoch metrics
-        np_epoch = {
-            "loss": np.stack(np_batch["loss"]).mean(),
-            "tag_ids": np.concatenate(
-                np_batch["tag_ids"]
-            ),  # shape: [epoch_size, seq_length]
-            "logits": np.concatenate(
-                np_batch["logits"]
-            ),  # shape: [epoch_size, seq_length, num_tags]
-        }
-
-        # epoch metrics
-        epoch_metrics, epoch_tags = self.compute_metrics(phase, np_epoch)
+        ner_model_evaluation = NerModelEvaluation(
+            current_epoch=self.current_epoch,
+            tag_list=self.tag_list,
+            dataset_tags=self.params.dataset_tags,
+            default_logger=self.default_logger,
+            logged_metrics=self.logged_metrics,
+        )
+        epoch_metrics, classification_report, epoch_loss = ner_model_evaluation.execute(
+            phase, outputs
+        )
+        self._log_metrics_and_classification_report(
+            phase, epoch_metrics, classification_report
+        )
+        return {f"{phase}_loss": epoch_loss}
 
+    def _log_metrics_and_classification_report(
+        self, phase: str, epoch_metrics: Dict[str, np.array], classification_report: str
+    ) -> None:
+        """
+        Args:
+            phase:                 [str] 'val', 'test'
+            epoch_metrics          [dict] w/ keys 'all_acc', 'fil_f1_micro', .. & values = [np array]
+            classification_report: [str]
+        """
         # tracked metrics & classification reports
-        self.add_epoch_metrics(
+        self._add_epoch_metrics(
             phase, self.current_epoch, epoch_metrics
         )  # attr: epoch_metrics
-        self.get_classification_report(
-            phase, self.current_epoch, epoch_tags
-        )  # attr: classification_reports
 
         # logging: tb
-        self.write_metrics_for_tensorboard(phase, epoch_metrics)
+        self._write_metrics_for_tensorboard(phase, epoch_metrics)
 
         # logging: mlflow
         if phase == "val":
             self.mlflow_client.log_metrics(self.current_epoch, epoch_metrics)
         self.mlflow_client.log_classification_report(
-            self.classification_reports[phase][self.current_epoch],
+            classification_report,
             overwrite=(phase == "val" and self.current_epoch == 0),
         )
         self.mlflow_client.finish_artifact_mlflow()
 
         # print
-        self._print_metrics(
-            phase, epoch_metrics, self.classification_reports[phase][self.current_epoch]
-        )
+        self._print_metrics(phase, epoch_metrics, classification_report)
 
         self.default_logger.log_debug(f"--> {phase}: epoch done")
 
-        return {f"{phase}_loss": np_epoch["loss"]}
-
-    def compute_metrics(self, phase, _np_dict):
-        """
-        computes loss, acc, f1 scores for size/phase = batch/train or epoch/val-test
-        ----------------------------------------------------------------------------
-        :param phase:          [str], 'train', 'val', 'test'
-        :param _np_dict:       [dict] w/ key-value pairs:
-                                     'loss':     [np value]
-                                     'tag_ids':  [np array] of shape [batch_size, seq_length]
-                                     'logits'    [np array] of shape [batch_size, seq_length, num_tags]
-        :return: metrics       [dict] w/ keys 'all+_loss', 'all+_acc', 'fil_f1_micro', .. & values = [np array]
-                 tags          [dict] w/ keys 'true', 'pred'      & values = [np array]
-        """
-        # batch / dataset
-        tag_ids = dict()
-        tag_ids["true"], tag_ids["pred"] = self._reduce_and_flatten(
-            _np_dict["tag_ids"], _np_dict["logits"]
-        )
-
-        tags = {
-            field: self._convert_tag_ids_to_tags(tag_ids[field])
-            for field in ["true", "pred"]
-        }
-        tags = self._get_rid_of_pad_tag_occurrences(tags)
-
-        self.default_logger.log_debug("phase:", phase)
-        self.default_logger.log_debug(
-            "true:", np.shape(tags["true"]), list(set(tags["true"]))
-        )
-        self.default_logger.log_debug(
-            "pred:", np.shape(tags["pred"]), list(set(tags["pred"]))
-        )
-
-        if phase == "val":
-            for field in ["true", "pred"]:
-                np.save(f'{env_variable("DIR_RESULTS")}/{field}.npy', tags[field])
-
-        # batch / dataset metrics
-        metrics = {"all+_loss": _np_dict["loss"]}
-        for tag_subset in [
-            "all+",
-            "all",
-            "fil",
-            "chk",
-        ] + self.tag_list:  # self._get_filtered_tags():
-            metrics.update(
-                self._compute_metrics_for_tags_subset(
-                    tags, phase, tag_subset=tag_subset
-                )
-            )
-
-        return metrics, tags
-
-    @staticmethod
-    def _reduce_and_flatten(_np_tag_ids, _np_logits):
-        """
-        helper method
-        reduce _np_logits (3D -> 2D), flatten both np arrays (2D -> 1D)
-        ---------------------------------------------------------------
-        :param _np_tag_ids: [np array] of shape [batch_size, seq_length]
-        :param _np_logits:  [np array] of shape [batch_size, seq_length, num_tags]
-        :return: true_flat: [np array] of shape [batch_size * seq_length], _np_tag_ids               flattened
-                 pred_flat: [np array] of shape [batch_size * seq_length], _np_logits    reduced and flattened
-        """
-        true_flat = _np_tag_ids.flatten()
-        pred_flat = np.argmax(_np_logits, axis=2).flatten()
-        return true_flat, pred_flat
-
-    def _convert_tag_ids_to_tags(self, _tag_ids):
-        """
-        helper method
-        convert tag_ids (int) to tags (str)
-        -----------------------------------
-        :param _tag_ids: [np array] of shape [batch_size * seq_length] with [int] elements
-        :return: _tags:  [np array] of shape [batch_size * seq_length] with [str] elements
-        """
-        return np.array([self.tag_list[elem] for elem in _tag_ids])
-
-    @staticmethod
-    def _get_rid_of_pad_tag_occurrences(_tags):
-        """
-        get rid of all elements where '[PAD]' occurs in true array
-        ----------------------------------------------------------
-        :param _tags:      [dict] w/ keys = 'true', 'pred' and
-                                     values = [np array] of shape [batch_size * seq_length]
-        :return: _tags_new [dict] w/ keys = 'true', 'pred' and
-                                     values = [np array] of shape [batch_size * seq_length - # of pad occurrences]
-        """
-        pad_indices = np.where(_tags["true"] == "[PAD]")
-        return {key: np.delete(_tags[key], pad_indices) for key in ["true", "pred"]}
-
-    def _compute_metrics_for_tags_subset(self, _tags, _phase, tag_subset: str):
-        """
-        helper method
-        compute metrics for tags subset (e.g. 'all', 'fil')
-        ---------------------------------------------------
-        :param _tags:      [dict] w/ keys 'true', 'pred'      & values = [np array]
-        :param _phase:     [str], 'train', 'val'
-        :param tag_subset: [str], e.g. 'all+', 'all', 'fil', 'PER'
-        :return: _metrics  [dict] w/ keys = metric (e.g. 'all_precision_micro') and value = [float]
-        """
-        tag_list = self._get_filtered_tags(tag_subset)
-        if tag_subset in ["all+", "all", "fil", "chk"]:
-            tag_group = [tag_subset]
-        else:
-            tag_group = ["ind"]
-        if tag_subset == "chk":
-            level = "chunk"
-        else:
-            level = "token"
-
-        ner_metrics = NerMetrics(
-            _tags["true"],
-            _tags["pred"],
-            tag_list=tag_list,
-            level=level,
-            plain_tags=self.params.dataset_tags == "plain",
-        )
-        ner_metrics.compute(
-            self.logged_metrics.get_metrics(tag_group=tag_group, phase_group=[_phase])
-        )
-        results = ner_metrics.results_as_dict()
-
-        _metrics = dict()
-        # simple
-        for metric_type in self.logged_metrics.get_metrics(
-            tag_group=tag_group,
-            phase_group=[_phase],
-            micro_macro_group=["simple"],
-            exclude=["loss"],
-        ):
-            # if results[metric_type] is not None:
-            _metrics[f"{tag_subset}_{metric_type}"] = results[metric_type]
-
-        # micro
-        for metric_type in self.logged_metrics.get_metrics(
-            tag_group=tag_group, phase_group=[_phase], micro_macro_group=["micro"]
-        ):
-            # if results[f'{metric_type}_micro'] is not None:
-            if tag_group == ["ind"]:
-                _metrics[f"{tag_subset}_{metric_type}"] = results[
-                    f"{metric_type}_micro"
-                ]
-            else:
-                _metrics[f"{tag_subset}_{metric_type}_micro"] = results[
-                    f"{metric_type}_micro"
-                ]
-
-        # macro
-        for metric_type in self.logged_metrics.get_metrics(
-            tag_group=tag_group, phase_group=[_phase], micro_macro_group=["macro"]
-        ):
-            # if results[f'{metric_type}_macro'] is not None:
-            _metrics[f"{tag_subset}_{metric_type}_macro"] = results[
-                f"{metric_type}_macro"
-            ]
-
-        return _metrics
-
-    def _get_filtered_tags(self, _tag_subset):
-        """
-        helper method
-        get list of filtered tags corresponding to _tag_subset name
-        -----------------------------------------------------------
-        :param _tag_subset: [str], e.g. 'all+', 'all', 'fil', 'PER'
-        :return: _filtered_tags: [list] of filtered tags [str]
-        """
-        if _tag_subset == "all++":
-            return None
-        elif _tag_subset == "all+":
-            return [tag for tag in self.tag_list if not tag == "[PAD]"]
-        elif _tag_subset == "all":
-            return [tag for tag in self.tag_list if not tag.startswith("[")]
-        elif _tag_subset in ["fil", "chk"]:
-            return [
-                tag for tag in self.tag_list if not (tag.startswith("[") or tag == "O")
-            ]
-        else:
-            return [_tag_subset]
-
-    def add_epoch_metrics(self, phase, epoch, _epoch_metrics):
+    def _add_epoch_metrics(self, phase, epoch, _epoch_metrics):
         """
         add _epoch_metrics to attribute/dict epoch_<phase>_metrics
         --------------------------------------------------------------
@@ -592,55 +426,7 @@ class NerModel(pl.LightningModule, ABC):
         """
         self.epoch_metrics[phase][epoch] = _epoch_metrics
 
-    def get_classification_report(self, phase, epoch, epoch_tags):
-        """
-        get token-based (sklearn) & chunk-based (seqeval) classification report
-        -----------------------------------------------------------------------
-        :param: epoch:                         [int]
-        :param: epoch_tags:                    [dict] w/ keys 'true', 'pred'      & values = [np array]
-        :changed attr: classification reports: [dict] w/ keys = epoch [int], values = classification report [str]
-        :return: -
-        """
-        warnings.filterwarnings("ignore")
-
-        self.classification_reports[phase][epoch] = ""
-
-        # token-based classification report
-        tag_list_filtered = self._get_filtered_tags("fil")
-        self.classification_reports[phase][
-            epoch
-        ] += f"\n>>> Phase: {phase} | Epoch: {epoch}"
-        self.classification_reports[phase][
-            epoch
-        ] += "\n--- token-based (sklearn) classification report on fil ---\n"
-        self.classification_reports[phase][epoch] += classification_report_sklearn(
-            epoch_tags["true"], epoch_tags["pred"], labels=tag_list_filtered
-        )
-
-        # chunk-based classification report
-        epoch_tags_chunk = dict()
-        for field in ["true", "pred"]:
-            epoch_tags_chunk[field] = convert_to_chunk(
-                epoch_tags[field], to_bio=self.params.dataset_tags == "plain"
-            )
-        self.default_logger.log_debug("> dataset_tags:", self.params.dataset_tags)
-        self.default_logger.log_debug(
-            "> epoch_tags_chunk[true]:", list(set(epoch_tags_chunk["true"]))
-        )
-        self.default_logger.log_debug(
-            "> epoch_tags_chunk[pred]:", list(set(epoch_tags_chunk["pred"]))
-        )
-
-        self.classification_reports[phase][
-            epoch
-        ] += "\n--- chunk-based (seqeval) classification report on fil ---\n"
-        self.classification_reports[phase][epoch] += classification_report_seqeval(
-            epoch_tags_chunk["true"], epoch_tags_chunk["pred"], suffix=False
-        )
-
-        warnings.resetwarnings()
-
-    def write_metrics_for_tensorboard(self, phase, metrics):
+    def _write_metrics_for_tensorboard(self, phase, metrics):
         """
         write metrics for tensorboard
         -----------------------------
@@ -702,16 +488,16 @@ class NerModel(pl.LightningModule, ABC):
             f"--- Epoch #{self.current_epoch} {phase.ljust(5).upper()} ----"
         )
         self.default_logger.log_info(
-            "all+ loss:         {:.2f}".format(_metrics["all+_loss"])
+            "all loss:         {:.2f}".format(_metrics["all_loss"])
         )
         self.default_logger.log_info(
-            "all+ acc:          {:.2f}".format(_metrics["all+_acc"])
+            "all acc:          {:.2f}".format(_metrics["all_acc"])
         )
         self.default_logger.log_debug(
-            "all+ f1 (micro):   {:.2f}".format(_metrics["all+_f1_micro"])
+            "all f1 (micro):   {:.2f}".format(_metrics["all_f1_micro"])
         )
         self.default_logger.log_debug(
-            "all+ f1 (macro):   {:.2f}".format(_metrics["all+_f1_macro"])
+            "all f1 (macro):   {:.2f}".format(_metrics["all_f1_macro"])
         )
         self.default_logger.log_info(
             "all  f1 (micro):   {:.2f}".format(_metrics["all_f1_micro"])
